# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import random
import re
import threading
from datetime import timedelta
from threading import Lock
from typing import Dict

import rocketmq
from google.protobuf.duration_pb2 import Duration
from rocketmq.client_config import ClientConfig
from rocketmq.consumer import Consumer
from rocketmq.definition import PermissionHelper
from rocketmq.filter_expression import FilterExpression
from rocketmq.log import logger
from rocketmq.message import MessageView
from rocketmq.protocol.definition_pb2 import Resource
from rocketmq.protocol.definition_pb2 import Resource as ProtoResource
from rocketmq.protocol.service_pb2 import \
    AckMessageEntry as ProtoAckMessageEntry
from rocketmq.protocol.service_pb2 import \
    AckMessageRequest as ProtoAckMessageRequest
from rocketmq.protocol.service_pb2 import \
    ChangeInvisibleDurationRequest as ProtoChangeInvisibleDurationRequest
from rocketmq.rpc_client import Endpoints
from rocketmq.session_credentials import (SessionCredentials,
                                          SessionCredentialsProvider)
from rocketmq.simple_subscription_settings import SimpleSubscriptionSettings
from rocketmq.state import State
from rocketmq.utils import get_positive_mod


class SubscriptionLoadBalancer:
    """This class serves as a load balancer for message subscription.
    It keeps track of a rotating index to help distribute the load evenly.
    """

    def __init__(self, topic_route_data):
        #: current index for message queue selection
        self._index = random.randint(0, 10000)  # assuming a range of 0-10000
        #: thread lock to ensure atomic update to the index
        self._index_lock = threading.Lock()

        #: filter the message queues which are readable and from the master broker
        self._message_queues = [
            mq for mq in topic_route_data.message_queues
            if PermissionHelper().is_readable(mq.permission)
            and mq.broker.id == rocketmq.utils.master_broker_id
        ]

    def update(self, topic_route_data):
        """Updates the message queues based on the new topic route data."""
        self._index += 1
        self._message_queues = [
            mq for mq in topic_route_data.message_queues
            if PermissionHelper().is_readable(mq.permission)
            and mq.broker.id == rocketmq.utils.master_broker_id
        ]
        return self

    def take_message_queue(self):
        """Fetches the next message queue based on the current index."""
        with self._index_lock:
            index = get_positive_mod(self._index, len(self._message_queues))
            self._index += 1
        return self._message_queues[index]


class SimpleConsumer(Consumer):
    """The SimpleConsumer class extends the Client class and is used to consume
    messages from specific topics in RocketMQ.
    """

    def __init__(self, client_config: ClientConfig, consumer_group: str, await_duration: int, subscription_expressions: Dict[str, FilterExpression]):
        """Create a new SimpleConsumer.

        :param client_config: The configuration for the client.
        :param consumer_group: The consumer group.
        :param await_duration: The await duration.
        :param subscription_expressions: The subscription expressions.
        """
        super().__init__(client_config, consumer_group)

        self._consumer_group = consumer_group
        self._await_duration = await_duration
        self._subscription_expressions = subscription_expressions

        self._simple_subscription_settings = SimpleSubscriptionSettings(self.client_id, self.endpoints, self._consumer_group, timedelta(seconds=10), 10, self._subscription_expressions)
        self._subscription_route_data_cache = {}
        self._topic_round_robin_index = 0
        self._state_lock = Lock()
        self._state = State.New
        self._subscription_load_balancer = {}  # A dictionary to keep subscription load balancers

    def get_topics(self):
        return set(self._subscription_expressions.keys())

    def get_settings(self):
        return self._simple_subscription_settings

    async def subscribe(self, topic: str, filter_expression: FilterExpression):
        if self._state != State.Running:
            raise Exception("Simple consumer is not running")

        await self.get_subscription_load_balancer(topic)
        self._subscription_expressions[topic] = filter_expression

    def unsubscribe(self, topic: str):
        if self._state != State.Running:
            raise Exception("Simple consumer is not running")
        try:
            self._subscription_expressions.pop(topic)
        except KeyError:
            pass

    async def start(self):
        """Start the RocketMQ consumer and log the operation."""
        logger.info(f"Begin to start the rocketmq consumer, client_id={self.client_id}")
        with self._state_lock:
            if self._state != State.New:
                raise Exception("Consumer already started")
            await super().start()
            # Start all necessary operations
            self._state = State.Running
        logger.info(f"The rocketmq consumer starts successfully, client_id={self.client_id}")

    async def shutdown(self):
        """Shutdown the RocketMQ consumer and log the operation."""
        logger.info(f"Begin to shutdown the rocketmq consumer, client_id={self.client_id}")
        with self._state_lock:
            if self._state != State.Running:
                raise Exception("Consumer is not running")
            # Shutdown all necessary operations
            self._state = State.Terminated
        await super().shutdown()
        logger.info(f"Shutdown the rocketmq consumer successfully, client_id={self.client_id}")

    def update_subscription_load_balancer(self, topic, topic_route_data):
        # if a load balancer for this topic already exists in the subscription routing data cache, update it
        subscription_load_balancer = self._subscription_route_data_cache.get(topic)
        if subscription_load_balancer:
            subscription_load_balancer.update(topic_route_data)
        # otherwise, create a new subscription load balancer
        else:
            subscription_load_balancer = SubscriptionLoadBalancer(topic_route_data)

        # store new or updated subscription load balancers in the subscription routing data cache
        self._subscription_route_data_cache[topic] = subscription_load_balancer
        return subscription_load_balancer

    async def get_subscription_load_balancer(self, topic):
        # if a load balancer for this topic already exists in the subscription routing data cache, return it
        subscription_load_balancer = self._subscription_route_data_cache.get(topic)
        if subscription_load_balancer:
            return subscription_load_balancer

        # otherwise, obtain the routing data for the topic
        topic_route_data = await self.get_route_data(topic)
        # update subscription load balancer
        return self.update_subscription_load_balancer(topic, topic_route_data)

    async def receive(self, max_message_num, invisible_duration):
        if self._state != State.Running:
            raise Exception("Simple consumer is not running")
        if max_message_num <= 0:
            raise Exception("maxMessageNum must be greater than 0")
        copy = dict(self._subscription_expressions)
        topics = list(copy.keys())
        if len(topics) == 0:
            raise ValueError("There is no topic to receive message")

        index = (self._topic_round_robin_index + 1) % len(topics)
        self._topic_round_robin_index = index
        topic = topics[index]
        filter_expression = self._subscription_expressions[topic]
        subscription_load_balancer = await self.get_subscription_load_balancer(topic)
        mq = subscription_load_balancer.take_message_queue()
        request = self.wrap_receive_message_request(max_message_num, mq, filter_expression, self._await_duration, invisible_duration)
        result = await self.receive_message(request, mq, self._await_duration)
        return result.messages

    def wrap_change_invisible_duration(self, message_view: MessageView, invisible_duration):
        topic_resource = ProtoResource()
        topic_resource.name = message_view.topic

        request = ProtoChangeInvisibleDurationRequest()
        request.topic.CopyFrom(topic_resource)
        group = ProtoResource()
        group.name = message_view.message_group
        logger.debug(message_view.message_group)
        request.group.CopyFrom(group)
        request.receipt_handle = message_view.receipt_handle
        request.invisible_duration.CopyFrom(Duration(seconds=invisible_duration))
        request.message_id = message_view.message_id

        return request

    async def change_invisible_duration(self, message_view: MessageView, invisible_duration):
        if self._state != State.Running:
            raise Exception("Simple consumer is not running")

        request = self.wrap_change_invisible_duration(message_view, invisible_duration)
        result = await self.client_manager.change_invisible_duration(
            message_view.message_queue.broker.endpoints,
            request,
            self.client_config.request_timeout
        )
        logger.debug(result)

    async def ack(self, message_view: MessageView):
        if self._state != State.Running:
            raise Exception("Simple consumer is not running")
        request = self.wrap_ack_message_request(message_view)
        result = await self.client_manager.ack_message(message_view.message_queue.broker.endpoints, request=request, timeout_seconds=self.client_config.request_timeout)
        logger.info(result)

    def get_protobuf_group(self):
        return ProtoResource(name=self.consumer_group)

    def wrap_ack_message_request(self, message_view: MessageView):
        topic_resource = ProtoResource()
        topic_resource.name = message_view.topic
        entry = ProtoAckMessageEntry()
        entry.message_id = message_view.message_id
        entry.receipt_handle = message_view.receipt_handle

        request = ProtoAckMessageRequest(group=self.get_protobuf_group(), topic=topic_resource, entries=[entry])
        return request

    class Builder:
        def __init__(self):
            self._consumer_group_regex = re.compile(r"^[%a-zA-Z0-9_-]+$")
            self._clientConfig = None
            self._consumerGroup = None
            self._awaitDuration = None
            self._subscriptionExpressions = {}

        def set_client_config(self, client_config: ClientConfig):
            if client_config is None:
                raise ValueError("clientConfig should not be null")
            self._clientConfig = client_config
            return self

        def set_consumer_group(self, consumer_group: str):
            if consumer_group is None:
                raise ValueError("consumerGroup should not be null")
            # Assuming CONSUMER_GROUP_REGEX is defined in the outer scope
            if not re.match(self._consumer_group_regex, consumer_group):
                raise ValueError(f"topic does not match the regex {self._consumer_group_regex}")
            self._consumerGroup = consumer_group
            return self

        def set_await_duration(self, await_duration: int):
            self._awaitDuration = await_duration
            return self

        def set_subscription_expression(self, subscription_expressions: Dict[str, FilterExpression]):
            if subscription_expressions is None:
                raise ValueError("subscriptionExpressions should not be null")
            if len(subscription_expressions) == 0:
                raise ValueError("subscriptionExpressions should not be empty")
            self._subscriptionExpressions = subscription_expressions
            return self

        async def build(self):
            if self._clientConfig is None:
                raise ValueError("clientConfig has not been set yet")
            if self._consumerGroup is None:
                raise ValueError("consumerGroup has not been set yet")
            if len(self._subscriptionExpressions) == 0:
                raise ValueError("subscriptionExpressions has not been set yet")

            simple_consumer = SimpleConsumer(self._clientConfig, self._consumerGroup, self._awaitDuration, self._subscriptionExpressions)
            await simple_consumer.start()
            return simple_consumer


async def test():
    credentials = SessionCredentials("username", "password")
    credentials_provider = SessionCredentialsProvider(credentials)
    client_config = ClientConfig(
        endpoints=Endpoints("endpoint"),
        session_credentials_provider=credentials_provider,
        ssl_enabled=True,
    )
    topic = Resource()
    topic.name = "normal_topic"

    consumer_group = "yourConsumerGroup"
    subscription = {topic.name: FilterExpression("*")}
    simple_consumer = (await SimpleConsumer.Builder()
                       .set_client_config(client_config)
                       .set_consumer_group(consumer_group)
                       .set_await_duration(15)
                       .set_subscription_expression(subscription)
                       .build())
    logger.info(simple_consumer)
    # while True:
    message_views = await simple_consumer.receive(16, 15)
    logger.info(message_views)
    for message in message_views:
        logger.info(message.body)
        logger.info(f"Received a message, topic={message.topic}, message-id={message.message_id}, body-size={len(message.body)}")
        await simple_consumer.ack(message)
        logger.info(f"Message is acknowledged successfully, message-id={message.message_id}")


async def test_fifo_message():
    credentials = SessionCredentials("username", "password")
    credentials_provider = SessionCredentialsProvider(credentials)
    client_config = ClientConfig(
        endpoints=Endpoints("endpoint"),
        session_credentials_provider=credentials_provider,
        ssl_enabled=True,
    )
    topic = Resource()
    topic.name = "fifo_topic"

    consumer_group = "yourConsumerGroup"
    subscription = {topic.name: FilterExpression("*")}
    simple_consumer = (await SimpleConsumer.Builder()
                       .set_client_config(client_config)
                       .set_consumer_group(consumer_group)
                       .set_await_duration(15)
                       .set_subscription_expression(subscription)
                       .build())
    logger.info(simple_consumer)
    # while True:
    message_views = await simple_consumer.receive(16, 15)
    # logger.info(message_views)
    for message in message_views:
        logger.info(message.body)
        logger.info(f"Received a message, topic={message.topic}, message-id={message.message_id}, body-size={len(message.body)}")
        await simple_consumer.ack(message)
        logger.info(f"Message is acknowledged successfully, message-id={message.message_id}")


async def test_change_invisible_duration():
    credentials = SessionCredentials("username", "password")
    credentials_provider = SessionCredentialsProvider(credentials)
    client_config = ClientConfig(
        endpoints=Endpoints("endpoint"),
        session_credentials_provider=credentials_provider,
        ssl_enabled=True,
    )
    topic = Resource()
    topic.name = "fifo_topic"

    consumer_group = "yourConsumerGroup"
    subscription = {topic.name: FilterExpression("*")}
    simple_consumer = (await SimpleConsumer.Builder()
                       .set_client_config(client_config)
                       .set_consumer_group(consumer_group)
                       .set_await_duration(15)
                       .set_subscription_expression(subscription)
                       .build())
    logger.info(simple_consumer)
    # while True:
    message_views = await simple_consumer.receive(16, 15)
    # logger.info(message_views)
    for message in message_views:
        await simple_consumer.change_invisible_duration(message_view=message, invisible_duration=3)
        logger.info(message.body)
        logger.info(f"Received a message, topic={message.topic}, message-id={message.message_id}, body-size={len(message.body)}")
        await simple_consumer.ack(message)
        logger.info(f"Message is acknowledged successfully, message-id={message.message_id}")


async def test_subscribe_unsubscribe():
    credentials = SessionCredentials("username", "password")
    credentials_provider = SessionCredentialsProvider(credentials)
    client_config = ClientConfig(
        endpoints=Endpoints("endpoint"),
        session_credentials_provider=credentials_provider,
        ssl_enabled=True,
    )
    topic = Resource()
    topic.name = "normal_topic"

    consumer_group = "yourConsumerGroup"
    subscription = {topic.name: FilterExpression("*")}
    simple_consumer = (await SimpleConsumer.Builder()
                       .set_client_config(client_config)
                       .set_consumer_group(consumer_group)
                       .set_await_duration(15)
                       .set_subscription_expression(subscription)
                       .build())
    logger.info(simple_consumer)
    # while True:
    message_views = await simple_consumer.receive(16, 15)
    logger.info(message_views)
    for message in message_views:
        logger.info(message.body)
        logger.info(f"Received a message, topic={message.topic}, message-id={message.message_id}, body-size={len(message.body)}")
        await simple_consumer.ack(message)
        logger.info(f"Message is acknowledged successfully, message-id={message.message_id}")
    simple_consumer.unsubscribe('normal_topic')
    await simple_consumer.subscribe('fifo_topic', FilterExpression("*"))
    message_views = await simple_consumer.receive(16, 15)
    logger.info(message_views)
    for message in message_views:
        logger.info(message.body)
        logger.info(f"Received a message, topic={message.topic}, message-id={message.message_id}, body-size={len(message.body)}")
        await simple_consumer.ack(message)
        logger.info(f"Message is acknowledged successfully, message-id={message.message_id}")

if __name__ == "__main__":
    asyncio.run(test_subscribe_unsubscribe())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import Optional

from rocketmq.protocol.definition_pb2 import ClientType as ProtoClientType
from rocketmq.protocol.definition_pb2 import RetryPolicy as ProtoRetryPolicy
from rocketmq.protocol.definition_pb2 import Settings as ProtoSettings
from rocketmq.rpc_client import Endpoints


class IRetryPolicy(ABC):
    @abstractmethod
    def GetMaxAttempts(self) -> int:
        pass

    @abstractmethod
    def GetNextAttemptDelay(self, attempt: int) -> datetime.timedelta:
        pass

    @abstractmethod
    def ToProtobuf(self) -> ProtoRetryPolicy:
        pass

    @abstractmethod
    def InheritBackoff(self, retryPolicy: ProtoRetryPolicy):
        pass


class ClientType(Enum):
    Producer = 1
    SimpleConsumer = 2
    PushConsumer = 3


class ClientTypeHelper:
    @staticmethod
    def to_protobuf(clientType):
        return {
            ClientType.Producer: ProtoClientType.PRODUCER,
            ClientType.SimpleConsumer: ProtoClientType.SIMPLE_CONSUMER,
            ClientType.PushConsumer: ProtoClientType.PUSH_CONSUMER,
        }.get(clientType, ProtoClientType.CLIENT_TYPE_UNSPECIFIED)


@dataclass
class Settings:
    ClientId: str
    ClientType: ClientType
    Endpoints: Endpoints
    RetryPolicy: Optional[IRetryPolicy]
    RequestTimeout: timedelta

    def to_protobuf(self):
        settings = ProtoSettings()
        return settings

    def Sync(self, settings):
        # Sync the settings properties from the Protobuf message
        pass

    def GetRetryPolicy(self):
        return self.RetryPolicy



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import socket
from typing import Dict

from rocketmq.exponential_backoff_retry_policy import \
    ExponentialBackoffRetryPolicy
from rocketmq.protocol.definition_pb2 import UA
from rocketmq.protocol.definition_pb2 import Publishing as ProtoPublishing
from rocketmq.protocol.definition_pb2 import Resource as ProtoResource
from rocketmq.protocol.definition_pb2 import Settings as ProtoSettings
from rocketmq.rpc_client import Endpoints
from rocketmq.settings import ClientType, ClientTypeHelper, Settings
from rocketmq.signature import Signature


class UserAgent:
    def __init__(self):
        self._version = Signature()._Signature__CLIENT_VERSION_KEY
        self._platform = platform.platform()
        self._hostname = socket.gethostname()

    def to_protobuf(self) -> UA:
        return UA(
            version=self._version, hostname=self._hostname, platform=self._platform
        )


class PublishingSettings(Settings):
    def __init__(
        self,
        client_id: str,
        endpoints: Endpoints,
        retry_policy: ExponentialBackoffRetryPolicy,
        request_timeout: int,
        topics: Dict[str, bool],
    ):
        super().__init__(
            client_id, ClientType.Producer, endpoints, retry_policy, request_timeout
        )
        self._max_body_size_bytes = 4 * 1024 * 1024
        self._validate_message_type = True
        self._topics = topics

    def get_max_body_size_bytes(self) -> int:
        return self._max_body_size_bytes

    def is_validate_message_type(self) -> bool:
        return self._validate_message_type

    def sync(self, settings: ProtoSettings) -> None:
        if settings.pub_sub_case != ProtoSettings.PubSubOneofCase.PUBLISHING:
            return

        self.retry_policy = self.retry_policy.inherit_backoff(settings.backoff_policy)
        self._validate_message_type = settings.publishing.validate_message_type
        self._max_body_size_bytes = settings.publishing.max_body_size

    def to_protobuf(self):
        topics = [ProtoResource(name=topic_name) for topic_name in self._topics]

        publishing = ProtoPublishing(
            topics=topics,
            validate_message_type=self._validate_message_type,
            max_body_size=self._max_body_size_bytes,
        )
        return ProtoSettings(
            publishing=publishing,
            access_point=self.Endpoints.to_protobuf(),
            client_type=ClientTypeHelper.to_protobuf(self.ClientType),
            user_agent=UserAgent().to_protobuf(),
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import threading

from rocketmq.protocol import definition_pb2, service_pb2
from rocketmq.protocol.definition_pb2 import Code as ProtoCode
from rocketmq.protocol.service_pb2 import HeartbeatRequest, QueryRouteRequest
from rocketmq.client_config import ClientConfig
from rocketmq.client_id_encoder import ClientIdEncoder
from rocketmq.definition import Resource, TopicRouteData
from rocketmq.log import logger
from rocketmq.rpc_client import Endpoints, RpcClient
from rocketmq.session import Session
from rocketmq.signature import Signature


class ScheduleWithFixedDelay:
    def __init__(self, action, delay, period):
        self.action = action
        self.delay = delay
        self.period = period
        self.task = None

    async def start(self):
        await asyncio.sleep(self.delay)
        while True:
            try:
                await self.action()
            except Exception as e:
                logger.error(e, "Failed to execute scheduled task")
            finally:
                await asyncio.sleep(self.period)

    def schedule(self):
        loop1 = asyncio.new_event_loop()
        asyncio.set_event_loop(loop1)
        self.task = asyncio.create_task(self.start())

    def cancel(self):
        if self.task:
            self.task.cancel()


class Client:
    """
    Main client class which handles interaction with the server.
    """
    def __init__(self, client_config: ClientConfig):
        """
        Initialization method for the Client class.

        :param client_config: Client configuration.
        :param topics: Set of topics that the client is subscribed to.
        """
        self.client_config = client_config
        self.client_id = ClientIdEncoder.generate()
        self.endpoints = client_config.endpoints

        #: A cache to store topic routes.
        self.topic_route_cache = {}

        #: A table to store session information.
        self.sessions_table = {}
        self.sessionsLock = threading.Lock()
        self.client_manager = ClientManager(self)

        #: A dictionary to store isolated items.
        self.isolated = dict()

    def get_topics(self):
        raise NotImplementedError("This method should be implemented by the subclass.")

    async def start(self):
        """
        Start method which initiates fetching of topic routes and schedules heartbeats.
        """
        # get topic route
        logger.debug(f"Begin to start the rocketmq client, client_id={self.client_id}")
        for topic in self.get_topics():
            self.topic_route_cache[topic] = await self.fetch_topic_route(topic)
        scheduler = ScheduleWithFixedDelay(self.heartbeat, 3, 12)
        scheduler_sync_settings = ScheduleWithFixedDelay(self.sync_settings, 3, 12)
        scheduler.schedule()
        scheduler_sync_settings.schedule()
        logger.debug(f"Start the rocketmq client successfully, client_id={self.client_id}")

    async def shutdown(self):
        logger.debug(f"Begin to shutdown rocketmq client, client_id={self.client_id}")

        logger.debug(f"Shutdown the rocketmq client successfully, client_id={self.client_id}")

    async def heartbeat(self):
        """
        Asynchronous method that sends a heartbeat to the server.
        """
        try:
            endpoints = self.get_total_route_endpoints()
            request = HeartbeatRequest()
            request.client_type = definition_pb2.PRODUCER
            topic = Resource()
            topic.name = "normal_topic"
            # Collect task into a map.
            for item in endpoints:
                try:

                    task = await self.client_manager.heartbeat(item, request, self.client_config.request_timeout)
                    code = task.status.code
                    if code == ProtoCode.OK:
                        logger.info(f"Send heartbeat successfully, endpoints={item}, client_id={self.client_id}")

                        if item in self.isolated:
                            self.isolated.pop(item)
                            logger.info(f"Rejoin endpoints which was isolated before, endpoints={item}, "
                                        + f"client_id={self.client_id}")
                        return
                    status_message = task.status.message
                    logger.info(f"Failed to send heartbeat, endpoints={item}, code={code}, "
                                + f"status_message={status_message}, client_id={self.client_id}")
                except Exception:
                    logger.error(f"Failed to send heartbeat, endpoints={item}")
        except Exception as e:
            logger.error(f"[Bug] unexpected exception raised during heartbeat, client_id={self.client_id}, Exception: {str(e)}")

    def get_total_route_endpoints(self):
        """
        Method that returns all route endpoints.
        """
        endpoints = set()
        for item in self.topic_route_cache.items():
            for endpoint in [mq.broker.endpoints for mq in item[1].message_queues]:
                endpoints.add(endpoint)
        return endpoints

    async def get_route_data(self, topic):
        """
        Asynchronous method that fetches route data for a given topic.

        :param topic: The topic to fetch route data for.
        """
        if topic in self.topic_route_cache:
            return self.topic_route_cache[topic]
        topic_route_data = await self.fetch_topic_route(topic=topic)
        return topic_route_data

    def get_client_config(self):
        """
        Method to return client configuration.
        """
        return self.client_config

    async def sync_settings(self):
        total_route_endpoints = self.get_total_route_endpoints()

        for endpoints in total_route_endpoints:
            created, session = await self.get_session(endpoints)
            await session.sync_settings(True)
            logger.info(f"Sync settings to remote, endpoints={endpoints}")

    def stats(self):
        # TODO: stats implement
        pass

    async def notify_client_termination(self):
        pass

    async def on_recover_orphaned_transaction_command(self, endpoints, command):
        pass

    async def on_verify_message_command(self, endpoints, command):
        logger.warn(f"Ignore verify message command from remote, which is not expected, clientId={self.client_id}, "
                    + f"endpoints={endpoints}, command={command}")
        pass

    async def on_print_thread_stack_trace_command(self, endpoints, command):
        pass

    async def on_settings_command(self, endpoints, settings):
        pass

    async def on_topic_route_data_fetched(self, topic, topic_route_data):
        """
        Asynchronous method that handles the process once the topic route data is fetched.

        :param topic: The topic for which the route data is fetched.
        :param topic_route_data: The fetched topic route data.
        """
        route_endpoints = set()
        for mq in topic_route_data.message_queues:
            route_endpoints.add(mq.broker.endpoints)

        existed_route_endpoints = self.get_total_route_endpoints()
        new_endpoints = route_endpoints.difference(existed_route_endpoints)

        for endpoints in new_endpoints:
            created, session = await self.get_session(endpoints)
            if not created:
                continue
            logger.info(f"Begin to establish session for endpoints={endpoints}, client_id={self.client_id}")
            await session.sync_settings(True)
            logger.info(f"Establish session for endpoints={endpoints} successfully, client_id={self.client_id}")

        self.topic_route_cache[topic] = topic_route_data

    async def fetch_topic_route0(self, topic):
        """
        Asynchronous method that fetches the topic route.

        :param topic: The topic to fetch the route for.
        """
        try:
            req = QueryRouteRequest()
            req.topic.name = topic
            address = req.endpoints.addresses.add()
            address.host = self.endpoints.Addresses[0].host
            address.port = self.endpoints.Addresses[0].port
            req.endpoints.scheme = self.endpoints.scheme.to_protobuf(self.endpoints.scheme)
            response = await self.client_manager.query_route(self.endpoints, req, 10)
            code = response.status.code
            if code != ProtoCode.OK:
                logger.error(f"Failed to fetch topic route, client_id={self.client_id}, topic={topic}, code={code}, "
                             + f"statusMessage={response.status.message}")
            message_queues = response.message_queues
            return TopicRouteData(message_queues)
        except Exception as e:
            logger.error(e, f"Failed to fetch topic route, client_id={self.client_id}, topic={topic}")
            raise

    async def fetch_topic_route(self, topic):
        """
        Asynchronous method that fetches the topic route and updates the data.

        :param topic: The topic to fetch the route for.
        """
        topic_route_data = await self.fetch_topic_route0(topic)
        await self.on_topic_route_data_fetched(topic, topic_route_data)
        logger.info(f"Fetch topic route successfully, client_id={self.client_id}, topic={topic}, topicRouteData={topic_route_data}")
        return topic_route_data

    async def get_session(self, endpoints):
        """
        Asynchronous method that gets the session for a given endpoint.

        :param endpoints: The endpoints to get the session for.
        """
        self.sessionsLock.acquire()
        try:
            # Session exists, return in advance.
            if endpoints in self.sessions_table:
                return (False, self.sessions_table[endpoints])
        finally:
            self.sessionsLock.release()

        self.sessionsLock.acquire()
        try:
            # Session exists, return in advance.
            if endpoints in self.sessions_table:
                return (False, self.sessions_table[endpoints])

            stream = self.client_manager.telemetry(endpoints, 10000000)
            created = Session(endpoints, stream, self)
            self.sessions_table[endpoints] = created
            return (True, created)
        finally:
            self.sessionsLock.release()

    def get_client_id(self):
        return self.client_id


class ClientManager:
    """Manager class for RPC Clients in a thread-safe manner.
    Each instance is created by a specific client and can manage
    multiple RPC clients.
    """

    def __init__(self, client: Client):
        #: The client that instantiated this manager.
        self.__client = client

        #: A dictionary that maps endpoints to the corresponding RPC clients.
        self.__rpc_clients = {}

        #: A lock used to ensure thread safety when accessing __rpc_clients.
        self.__rpc_clients_lock = threading.Lock()

    def __get_rpc_client(self, endpoints: Endpoints, ssl_enabled: bool):
        """Retrieve the RPC client corresponding to the given endpoints.
        If not present, a new RPC client is created and stored in __rpc_clients.

        :param endpoints: The endpoints associated with the RPC client.
        :param ssl_enabled: A flag indicating whether SSL is enabled.
        :return: The RPC client associated with the given endpoints.
        """
        with self.__rpc_clients_lock:
            rpc_client = self.__rpc_clients.get(endpoints)
            if rpc_client:
                return rpc_client
            rpc_client = RpcClient(endpoints.grpc_target(True), ssl_enabled)
            self.__rpc_clients[endpoints] = rpc_client
            return rpc_client

    async def query_route(
        self,
        endpoints: Endpoints,
        request: service_pb2.QueryRouteRequest,
        timeout_seconds: int,
    ):
        """Query the routing information.

        :param endpoints: The endpoints to query.
        :param request: The request containing the details of the query.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the query.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.query_route(request, metadata, timeout_seconds)

    async def heartbeat(
        self,
        endpoints: Endpoints,
        request: service_pb2.HeartbeatRequest,
        timeout_seconds: int,
    ):
        """Send a heartbeat to the server to indicate that the client is still alive.

        :param endpoints: The endpoints to send the heartbeat to.
        :param request: The request containing the details of the heartbeat.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the heartbeat.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.heartbeat(request, metadata, timeout_seconds)

    async def send_message(
        self,
        endpoints: Endpoints,
        request: service_pb2.SendMessageRequest,
        timeout_seconds: int,
    ):
        """Send a message to the server.

        :param endpoints: The endpoints to send the message to.
        :param request: The request containing the details of the message.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the message sending operation.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.send_message(request, metadata, timeout_seconds)

    async def query_assignment(
        self,
        endpoints: Endpoints,
        request: service_pb2.QueryAssignmentRequest,
        timeout_seconds: int,
    ):
        """Query the assignment information.

        :param endpoints: The endpoints to query.
        :param request: The request containing the details of the query.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the query.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.query_assignment(request, metadata, timeout_seconds)

    async def ack_message(
        self,
        endpoints: Endpoints,
        request: service_pb2.AckMessageRequest,
        timeout_seconds: int,
    ):
        """Send an acknowledgment for a message to the server.

        :param endpoints: The endpoints to send the acknowledgment to.
        :param request: The request containing the details of the acknowledgment.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the acknowledgment.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.ack_message(request, metadata, timeout_seconds)

    async def forward_message_to_dead_letter_queue(
        self,
        endpoints: Endpoints,
        request: service_pb2.ForwardMessageToDeadLetterQueueRequest,
        timeout_seconds: int,
    ):
        """Forward a message to the dead letter queue.

        :param endpoints: The endpoints to send the request to.
        :param request: The request containing the details of the message to forward.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the forward operation.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.forward_message_to_dead_letter_queue(
            request, metadata, timeout_seconds
        )

    async def end_transaction(
        self,
        endpoints: Endpoints,
        request: service_pb2.EndTransactionRequest,
        timeout_seconds: int,
    ):
        """Ends a transaction.

        :param endpoints: The endpoints to send the request to.
        :param request: The request to end the transaction.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the end transaction operation.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.end_transaction(request, metadata, timeout_seconds)

    async def notify_client_termination(
        self,
        endpoints: Endpoints,
        request: service_pb2.NotifyClientTerminationRequest,
        timeout_seconds: int,
    ):
        """Notify server about client termination.

        :param endpoints: The endpoints to send the notification to.
        :param request: The request containing the details of the termination.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the notification operation.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.notify_client_termination(
            request, metadata, timeout_seconds
        )

    async def change_invisible_duration(
        self,
        endpoints: Endpoints,
        request: service_pb2.ChangeInvisibleDurationRequest,
        timeout_seconds: int,
    ):
        """Change the invisible duration of a message.

        :param endpoints: The endpoints to send the request to.
        :param request: The request containing the new invisible duration.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The result of the change operation.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return await rpc_client.change_invisible_duration(
            request, metadata, timeout_seconds
        )

    async def receive_message(
        self,
        endpoints: Endpoints,
        request: service_pb2.ReceiveMessageRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)

        response = await rpc_client.receive_message(
            request, metadata, timeout_seconds
        )
        return response

    def telemetry(
        self,
        endpoints: Endpoints,
        timeout_seconds: int,
    ):
        """Fetch telemetry information.

        :param endpoints: The endpoints to send the request to.
        :param timeout_seconds: The maximum time to wait for a response.
        :return: The telemetry information.
        """
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        metadata = Signature.sign(self.__client.client_config, self.__client.client_id)
        return rpc_client.telemetry(metadata, timeout_seconds)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from rocketmq.protocol import service_pb2
from rocketmq.client import Client
from rocketmq.rpc_client import Endpoints, RpcClient


class ClientManager:
    def __init__(self, client: Client):
        self.__client = client
        self.__rpc_clients = {}
        self.__rpc_clients_lock = threading.Lock()

    def __get_rpc_client(self, endpoints: Endpoints, ssl_enabled: bool):
        with self.__rpc_clients_lock:
            rpc_client = self.__rpc_clients.get(endpoints)
            if rpc_client:
                return rpc_client
            rpc_client = RpcClient(endpoints, ssl_enabled)
            self.__rpc_clients[endpoints] = rpc_client
            return rpc_client

    async def query_route(
        self,
        endpoints: Endpoints,
        request: service_pb2.QueryRouteRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.query_route(request, timeout_seconds)

    async def heartbeat(
        self,
        endpoints: Endpoints,
        request: service_pb2.HeartbeatRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.heartbeat(request, timeout_seconds)

    async def send_message(
        self,
        endpoints: Endpoints,
        request: service_pb2.SendMessageRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.send_message(request, timeout_seconds)

    async def query_assignment(
        self,
        endpoints: Endpoints,
        request: service_pb2.QueryAssignmentRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.query_assignment(request, timeout_seconds)

    async def ack_message(
        self,
        endpoints: Endpoints,
        request: service_pb2.AckMessageRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.ack_message(request, timeout_seconds)

    async def forward_message_to_dead_letter_queue(
        self,
        endpoints: Endpoints,
        request: service_pb2.ForwardMessageToDeadLetterQueueRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.forward_message_to_dead_letter_queue(
            request, timeout_seconds
        )

    async def end_transaction(
        self,
        endpoints: Endpoints,
        request: service_pb2.EndTransactionRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.end_transaction(request, timeout_seconds)

    async def notify_client_termination(
        self,
        endpoints: Endpoints,
        request: service_pb2.NotifyClientTerminationRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.notify_client_termination(request, timeout_seconds)

    async def change_invisible_duration(
        self,
        endpoints: Endpoints,
        request: service_pb2.ChangeInvisibleDurationRequest,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.change_invisible_duration(request, timeout_seconds)

    async def telemetry(
        self,
        endpoints: Endpoints,
        request: service_pb2.TelemetryCommand,
        timeout_seconds: int,
    ):
        rpc_client = self.__get_rpc_client(
            endpoints, self.__client.client_config.ssl_enabled
        )
        return await rpc_client.telemetry()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

from rocketmq.log import logger
from rocketmq.protocol.service_pb2 import \
    TelemetryCommand as ProtoTelemetryCommand


class Session:
    def __init__(self, endpoints, streaming_call, client):
        self._endpoints = endpoints
        self._semaphore = asyncio.Semaphore(1)
        self._streaming_call = streaming_call
        self._client = client
        asyncio.create_task(self.loop())

    async def loop(self):
        try:
            while True:
                await self._streaming_call.read()
        except asyncio.exceptions.InvalidStateError as e:
            logger.error('Error:', e)

    async def write_async(self, telemetry_command: ProtoTelemetryCommand):
        await asyncio.sleep(1)
        try:
            await self._streaming_call.write(telemetry_command)
            # TODO handle read operation exceed the time limit
            # await asyncio.wait_for(self._streaming_call.read(), timeout=5)
        except asyncio.exceptions.InvalidStateError as e:
            self.on_error(e)
        except asyncio.TimeoutError:
            logger.error('Timeout: The read operation exceeded the time limit')

    async def sync_settings(self, await_resp):
        await self._semaphore.acquire()
        try:
            settings = self._client.get_settings()
            telemetry_command = ProtoTelemetryCommand()
            telemetry_command.settings.CopyFrom(settings.to_protobuf())
            await self.write_async(telemetry_command)
        finally:
            self._semaphore.release()

    def rebuild_telemetry(self):
        logger.info("Try to rebuild telemetry")
        stream = self._client.client_manager.telemetry(self._endpoints, 10)
        self._streaming_call = stream

    def on_error(self, exception):
        client_id = self._client.get_client_id()
        logger.error("Caught InvalidStateError: RPC already finished.")
        logger.error(f"Exception raised from stream, clientId={client_id}, endpoints={self._endpoints}", exception)
        max_retry = 3
        for i in range(max_retry):
            try:
                self.rebuild_telemetry()
                break
            except Exception as e:
                logger.error(f"An error occurred during rebuilding telemetry: {e}, attempt {i + 1} of {max_retry}")



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import os
import threading
import time
import uuid
from datetime import datetime, timezone


class MessageIdCodec:
    __MESSAGE_ID_VERSION_V1 = "01"

    @staticmethod
    def __get_process_fixed_string():
        mac = uuid.getnode()
        mac = format(mac, "012x")
        mac_bytes = bytes.fromhex(mac[-12:])
        pid = os.getpid() % 65536
        pid_bytes = pid.to_bytes(2, "big")
        return mac_bytes.hex().upper() + pid_bytes.hex().upper()

    @staticmethod
    def __get_seconds_since_custom_epoch():
        custom_epoch = datetime(2021, 1, 1, tzinfo=timezone.utc)
        now = datetime.now(timezone.utc)
        return int((now - custom_epoch).total_seconds())

    __PROCESS_FIXED_STRING_V1 = __get_process_fixed_string()
    __SECONDS_SINCE_CUSTOM_EPOCH = __get_seconds_since_custom_epoch()
    __SECONDS_START_TIMESTAMP = int(time.time())

    @staticmethod
    def __delta_seconds():
        return (
            int(time.time())
            - MessageIdCodec.__SECONDS_START_TIMESTAMP
            + MessageIdCodec.__SECONDS_SINCE_CUSTOM_EPOCH
        )

    @staticmethod
    def __int_to_bytes_with_big_endian(number: int, min_bytes: int):
        num_bytes = max(math.ceil(number.bit_length() / 8), min_bytes)
        return number.to_bytes(num_bytes, "big")

    __SEQUENCE = 0
    __SEQUENCE_LOCK = threading.Lock()

    @staticmethod
    def __get_and_increment_sequence():
        with MessageIdCodec.__SEQUENCE_LOCK:
            temp = MessageIdCodec.__SEQUENCE
            MessageIdCodec.__SEQUENCE += 1
            return temp

    @staticmethod
    def next_message_id():
        seconds = MessageIdCodec.__delta_seconds()
        seconds_bytes = MessageIdCodec.__int_to_bytes_with_big_endian(seconds, 4)[-4:]
        sequence_bytes = MessageIdCodec.__int_to_bytes_with_big_endian(
            MessageIdCodec.__get_and_increment_sequence(), 4
        )[-4:]
        return (
            MessageIdCodec.__MESSAGE_ID_VERSION_V1
            + MessageIdCodec.__PROCESS_FIXED_STRING_V1
            + seconds_bytes.hex().upper()
            + sequence_bytes.hex().upper()
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rocketmq.log import logger
from rocketmq.message import Message
from rocketmq.protocol.definition_pb2 import Code as ProtoCode
from rocketmq.protocol.definition_pb2 import Message as ProtoMessage
from rocketmq.protocol.definition_pb2 import Status as ProtoStatus
from rocketmq.protocol.service_pb2 import \
    ReceiveMessageRequest as ProtoReceiveMessageRequest


class RocketMQException(Exception):
    def __init__(self, status_code, request_id, status_message):
        self.status_code = status_code
        self.request_id = request_id
        self.status_message = status_message

    def __str__(self):
        return f"{self.__class__.__name__}: code={self.status_code}, requestId={self.request_id}, message={self.status_message}"


class BadRequestException(RocketMQException):
    pass


class UnauthorizedException(RocketMQException):
    pass


class PaymentRequiredException(RocketMQException):
    pass


class ForbiddenException(RocketMQException):
    pass


class NotFoundException(RocketMQException):
    pass


class PayloadTooLargeException(RocketMQException):
    pass


class TooManyRequestsException(RocketMQException):
    pass


class RequestHeaderFieldsTooLargeException(RocketMQException):
    pass


class InternalErrorException(RocketMQException):
    pass


class ProxyTimeoutException(RocketMQException):
    pass


class UnsupportedException(RocketMQException):
    pass


class StatusChecker:
    @staticmethod
    def check(status: ProtoStatus, request: Message, request_id: str):
        """Check the status of a request and raise an exception if necessary.

        :param status: A ProtoStatus object that contains the status code and message.
        :param request: The request message object.
        :param request_id: The ID of the request.
        :raise BadRequestException: If the status code indicates a bad request.
        :raise UnauthorizedException: If the status code indicates an unauthorized request.
        :raise PaymentRequiredException: If the status code indicates payment is required.
        :raise ForbiddenException: If the status code indicates a forbidden request.
        :raise NotFoundException: If the status code indicates a resource is not found.
        :raise PayloadTooLargeException: If the status code indicates the request payload is too large.
        :raise TooManyRequestsException: If the status code indicates too many requests.
        :raise RequestHeaderFieldsTooLargeException: If the status code indicates the request headers are too large.
        :raise InternalErrorException: If the status code indicates an internal error.
        :raise ProxyTimeoutException: If the status code indicates a proxy timeout.
        :raise UnsupportedException: If the status code indicates an unsupported operation.
        """
        status_code = status.code
        status_message = status.message

        if status_code in [ProtoCode.OK, ProtoCode.MULTIPLE_RESULTS]:
            return
        elif status_code in [
            ProtoCode.BAD_REQUEST,
            ProtoCode.ILLEGAL_ACCESS_POINT,
            ProtoCode.ILLEGAL_TOPIC,
            ProtoCode.ILLEGAL_CONSUMER_GROUP,
            ProtoCode.ILLEGAL_MESSAGE_TAG,
            ProtoCode.ILLEGAL_MESSAGE_KEY,
            ProtoCode.ILLEGAL_MESSAGE_GROUP,
            ProtoCode.ILLEGAL_MESSAGE_PROPERTY_KEY,
            ProtoCode.INVALID_TRANSACTION_ID,
            ProtoCode.ILLEGAL_MESSAGE_ID,
            ProtoCode.ILLEGAL_FILTER_EXPRESSION,
            ProtoCode.ILLEGAL_INVISIBLE_TIME,
            ProtoCode.ILLEGAL_DELIVERY_TIME,
            ProtoCode.INVALID_RECEIPT_HANDLE,
            ProtoCode.MESSAGE_PROPERTY_CONFLICT_WITH_TYPE,
            ProtoCode.UNRECOGNIZED_CLIENT_TYPE,
            ProtoCode.MESSAGE_CORRUPTED,
            ProtoCode.CLIENT_ID_REQUIRED,
            ProtoCode.ILLEGAL_POLLING_TIME,
        ]:
            raise BadRequestException(status_code, request_id, status_message)
        elif status_code == ProtoCode.UNAUTHORIZED:
            raise UnauthorizedException(status_code, request_id, status_message)
        elif status_code == ProtoCode.PAYMENT_REQUIRED:
            raise PaymentRequiredException(status_code, request_id, status_message)
        elif status_code == ProtoCode.FORBIDDEN:
            raise ForbiddenException(status_code, request_id, status_message)
        elif status_code == ProtoCode.MESSAGE_NOT_FOUND:
            if isinstance(request, ProtoReceiveMessageRequest):
                return
            else:
                # Fall through on purpose.
                status_code = ProtoCode.NOT_FOUND
        if status_code in [
            ProtoCode.NOT_FOUND,
            ProtoCode.TOPIC_NOT_FOUND,
            ProtoCode.CONSUMER_GROUP_NOT_FOUND,
        ]:
            raise NotFoundException(status_code, request_id, status_message)
        elif status_code in [
            ProtoCode.PAYLOAD_TOO_LARGE,
            ProtoCode.MESSAGE_BODY_TOO_LARGE,
        ]:
            raise PayloadTooLargeException(status_code, request_id, status_message)
        elif status_code == ProtoCode.TOO_MANY_REQUESTS:
            raise TooManyRequestsException(status_code, request_id, status_message)
        elif status_code in [
            ProtoCode.REQUEST_HEADER_FIELDS_TOO_LARGE,
            ProtoCode.MESSAGE_PROPERTIES_TOO_LARGE,
        ]:
            raise RequestHeaderFieldsTooLargeException(status_code, request_id, status_message)
        elif status_code in [
            ProtoCode.INTERNAL_ERROR,
            ProtoCode.INTERNAL_SERVER_ERROR,
            ProtoCode.HA_NOT_AVAILABLE,
        ]:
            raise InternalErrorException(status_code, request_id, status_message)
        elif status_code in [
            ProtoCode.PROXY_TIMEOUT,
            ProtoCode.MASTER_PERSISTENCE_TIMEOUT,
            ProtoCode.SLAVE_PERSISTENCE_TIMEOUT,
        ]:
            raise ProxyTimeoutException(status_code, request_id, status_message)
        elif status_code in [
            ProtoCode.UNSUPPORTED,
            ProtoCode.VERSION_UNSUPPORTED,
            ProtoCode.VERIFY_FIFO_MESSAGE_UNSUPPORTED,
        ]:
            raise UnsupportedException(status_code, request_id, status_message)
        else:
            logger.warning(f"Unrecognized status code={status_code}, requestId={request_id}, statusMessage={status_message}")
            raise UnsupportedException(status_code, request_id, status_message)


def main():
    # 创建一个表示'OK'状态的ProtoStatus
    status_ok = ProtoStatus()
    status_ok.code = ProtoCode.OK
    status_ok.message = "Everything is OK"

    # 创建一个表示'BadRequest'状态的ProtoStatus
    status_bad_request = ProtoStatus()
    status_bad_request.code = ProtoCode.BAD_REQUEST
    status_bad_request.message = "Bad request"

    # 创建一个表示'Unauthorized'状态的ProtoStatus
    status_unauthorized = ProtoStatus()
    status_unauthorized.code = ProtoCode.UNAUTHORIZED
    status_unauthorized.message = "Unauthorized"

    request = ProtoMessage()

    # 进行一些测试
    StatusChecker.check(status_ok, request, "request1")  # 不应抛出异常

    try:
        StatusChecker.check(status_bad_request, request, "request2")
    except BadRequestException as e:
        logger.error(f"Caught expected exception: {e}")

    try:
        StatusChecker.check(status_unauthorized, request, "request3")
    except UnauthorizedException as e:
        logger.error(f"Caught expected exception: {e}")


if __name__ == "__main__":
    main()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import socket
import threading
import time

import rocketmq.utils


class ClientIdEncoder:
    """This class generates a unique client ID for each client based on
    hostname, process id, index and the monotonic clock time.
    """

    #: The current index for client id generation.
    __INDEX = 0

    #: The lock used for thread-safe incrementing of the index.
    __INDEX_LOCK = threading.Lock()

    #: The separator used in the client id string.
    __CLIENT_ID_SEPARATOR = "@"

    @staticmethod
    def __get_and_increment_sequence() -> int:
        """Increment and return the current index in a thread-safe manner.

        :return: the current index after incrementing it.
        """
        with ClientIdEncoder.__INDEX_LOCK:
            temp = ClientIdEncoder.__INDEX
            ClientIdEncoder.__INDEX += 1
            return temp

    @staticmethod
    def generate() -> str:
        """Generate a unique client ID.

        :return: the generated client id
        """
        index = ClientIdEncoder.__get_and_increment_sequence()
        return (
            socket.gethostname()
            + ClientIdEncoder.__CLIENT_ID_SEPARATOR
            + str(os.getpid())
            + ClientIdEncoder.__CLIENT_ID_SEPARATOR
            + str(index)
            + ClientIdEncoder.__CLIENT_ID_SEPARATOR
            + str(rocketmq.utils.number_to_base(time.monotonic_ns(), 36))
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from protocol import definition_pb2 as _definition_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AckMessageEntry(_message.Message):
    __slots__ = ["message_id", "receipt_handle"]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    RECEIPT_HANDLE_FIELD_NUMBER: _ClassVar[int]
    message_id: str
    receipt_handle: str
    def __init__(self, message_id: _Optional[str] = ..., receipt_handle: _Optional[str] = ...) -> None: ...

class AckMessageRequest(_message.Message):
    __slots__ = ["entries", "group", "topic"]
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    entries: _containers.RepeatedCompositeFieldContainer[AckMessageEntry]
    group: _definition_pb2.Resource
    topic: _definition_pb2.Resource
    def __init__(self, group: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., topic: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., entries: _Optional[_Iterable[_Union[AckMessageEntry, _Mapping]]] = ...) -> None: ...

class AckMessageResponse(_message.Message):
    __slots__ = ["entries", "status"]
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    entries: _containers.RepeatedCompositeFieldContainer[AckMessageResultEntry]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., entries: _Optional[_Iterable[_Union[AckMessageResultEntry, _Mapping]]] = ...) -> None: ...

class AckMessageResultEntry(_message.Message):
    __slots__ = ["message_id", "receipt_handle", "status"]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    RECEIPT_HANDLE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    message_id: str
    receipt_handle: str
    status: _definition_pb2.Status
    def __init__(self, message_id: _Optional[str] = ..., receipt_handle: _Optional[str] = ..., status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ...) -> None: ...

class ChangeInvisibleDurationRequest(_message.Message):
    __slots__ = ["group", "invisible_duration", "message_id", "receipt_handle", "topic"]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    INVISIBLE_DURATION_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    RECEIPT_HANDLE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    group: _definition_pb2.Resource
    invisible_duration: _duration_pb2.Duration
    message_id: str
    receipt_handle: str
    topic: _definition_pb2.Resource
    def __init__(self, group: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., topic: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., receipt_handle: _Optional[str] = ..., invisible_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., message_id: _Optional[str] = ...) -> None: ...

class ChangeInvisibleDurationResponse(_message.Message):
    __slots__ = ["receipt_handle", "status"]
    RECEIPT_HANDLE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    receipt_handle: str
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., receipt_handle: _Optional[str] = ...) -> None: ...

class EndTransactionRequest(_message.Message):
    __slots__ = ["message_id", "resolution", "source", "topic", "trace_context", "transaction_id"]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    RESOLUTION_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    TRACE_CONTEXT_FIELD_NUMBER: _ClassVar[int]
    TRANSACTION_ID_FIELD_NUMBER: _ClassVar[int]
    message_id: str
    resolution: _definition_pb2.TransactionResolution
    source: _definition_pb2.TransactionSource
    topic: _definition_pb2.Resource
    trace_context: str
    transaction_id: str
    def __init__(self, topic: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., message_id: _Optional[str] = ..., transaction_id: _Optional[str] = ..., resolution: _Optional[_Union[_definition_pb2.TransactionResolution, str]] = ..., source: _Optional[_Union[_definition_pb2.TransactionSource, str]] = ..., trace_context: _Optional[str] = ...) -> None: ...

class EndTransactionResponse(_message.Message):
    __slots__ = ["status"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ...) -> None: ...

class ForwardMessageToDeadLetterQueueRequest(_message.Message):
    __slots__ = ["delivery_attempt", "group", "max_delivery_attempts", "message_id", "receipt_handle", "topic"]
    DELIVERY_ATTEMPT_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    MAX_DELIVERY_ATTEMPTS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    RECEIPT_HANDLE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    delivery_attempt: int
    group: _definition_pb2.Resource
    max_delivery_attempts: int
    message_id: str
    receipt_handle: str
    topic: _definition_pb2.Resource
    def __init__(self, group: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., topic: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., receipt_handle: _Optional[str] = ..., message_id: _Optional[str] = ..., delivery_attempt: _Optional[int] = ..., max_delivery_attempts: _Optional[int] = ...) -> None: ...

class ForwardMessageToDeadLetterQueueResponse(_message.Message):
    __slots__ = ["status"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ...) -> None: ...

class HeartbeatRequest(_message.Message):
    __slots__ = ["client_type", "group"]
    CLIENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    client_type: _definition_pb2.ClientType
    group: _definition_pb2.Resource
    def __init__(self, group: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., client_type: _Optional[_Union[_definition_pb2.ClientType, str]] = ...) -> None: ...

class HeartbeatResponse(_message.Message):
    __slots__ = ["status"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ...) -> None: ...

class NotifyClientTerminationRequest(_message.Message):
    __slots__ = ["group"]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    group: _definition_pb2.Resource
    def __init__(self, group: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ...) -> None: ...

class NotifyClientTerminationResponse(_message.Message):
    __slots__ = ["status"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ...) -> None: ...

class PrintThreadStackTraceCommand(_message.Message):
    __slots__ = ["nonce"]
    NONCE_FIELD_NUMBER: _ClassVar[int]
    nonce: str
    def __init__(self, nonce: _Optional[str] = ...) -> None: ...

class QueryAssignmentRequest(_message.Message):
    __slots__ = ["endpoints", "group", "topic"]
    ENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    endpoints: _definition_pb2.Endpoints
    group: _definition_pb2.Resource
    topic: _definition_pb2.Resource
    def __init__(self, topic: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., group: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., endpoints: _Optional[_Union[_definition_pb2.Endpoints, _Mapping]] = ...) -> None: ...

class QueryAssignmentResponse(_message.Message):
    __slots__ = ["assignments", "status"]
    ASSIGNMENTS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    assignments: _containers.RepeatedCompositeFieldContainer[_definition_pb2.Assignment]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., assignments: _Optional[_Iterable[_Union[_definition_pb2.Assignment, _Mapping]]] = ...) -> None: ...

class QueryRouteRequest(_message.Message):
    __slots__ = ["endpoints", "topic"]
    ENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    endpoints: _definition_pb2.Endpoints
    topic: _definition_pb2.Resource
    def __init__(self, topic: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., endpoints: _Optional[_Union[_definition_pb2.Endpoints, _Mapping]] = ...) -> None: ...

class QueryRouteResponse(_message.Message):
    __slots__ = ["message_queues", "status"]
    MESSAGE_QUEUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    message_queues: _containers.RepeatedCompositeFieldContainer[_definition_pb2.MessageQueue]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., message_queues: _Optional[_Iterable[_Union[_definition_pb2.MessageQueue, _Mapping]]] = ...) -> None: ...

class ReceiveMessageRequest(_message.Message):
    __slots__ = ["auto_renew", "batch_size", "filter_expression", "group", "invisible_duration", "long_polling_timeout", "message_queue"]
    AUTO_RENEW_FIELD_NUMBER: _ClassVar[int]
    BATCH_SIZE_FIELD_NUMBER: _ClassVar[int]
    FILTER_EXPRESSION_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    INVISIBLE_DURATION_FIELD_NUMBER: _ClassVar[int]
    LONG_POLLING_TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_QUEUE_FIELD_NUMBER: _ClassVar[int]
    auto_renew: bool
    batch_size: int
    filter_expression: _definition_pb2.FilterExpression
    group: _definition_pb2.Resource
    invisible_duration: _duration_pb2.Duration
    long_polling_timeout: _duration_pb2.Duration
    message_queue: _definition_pb2.MessageQueue
    def __init__(self, group: _Optional[_Union[_definition_pb2.Resource, _Mapping]] = ..., message_queue: _Optional[_Union[_definition_pb2.MessageQueue, _Mapping]] = ..., filter_expression: _Optional[_Union[_definition_pb2.FilterExpression, _Mapping]] = ..., batch_size: _Optional[int] = ..., invisible_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., auto_renew: bool = ..., long_polling_timeout: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class ReceiveMessageResponse(_message.Message):
    __slots__ = ["delivery_timestamp", "message", "status"]
    DELIVERY_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    delivery_timestamp: _timestamp_pb2.Timestamp
    message: _definition_pb2.Message
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., message: _Optional[_Union[_definition_pb2.Message, _Mapping]] = ..., delivery_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class RecoverOrphanedTransactionCommand(_message.Message):
    __slots__ = ["message", "transaction_id"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    TRANSACTION_ID_FIELD_NUMBER: _ClassVar[int]
    message: _definition_pb2.Message
    transaction_id: str
    def __init__(self, message: _Optional[_Union[_definition_pb2.Message, _Mapping]] = ..., transaction_id: _Optional[str] = ...) -> None: ...

class SendMessageRequest(_message.Message):
    __slots__ = ["messages"]
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    messages: _containers.RepeatedCompositeFieldContainer[_definition_pb2.Message]
    def __init__(self, messages: _Optional[_Iterable[_Union[_definition_pb2.Message, _Mapping]]] = ...) -> None: ...

class SendMessageResponse(_message.Message):
    __slots__ = ["entries", "status"]
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    entries: _containers.RepeatedCompositeFieldContainer[SendResultEntry]
    status: _definition_pb2.Status
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., entries: _Optional[_Iterable[_Union[SendResultEntry, _Mapping]]] = ...) -> None: ...

class SendResultEntry(_message.Message):
    __slots__ = ["message_id", "offset", "status", "transaction_id"]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TRANSACTION_ID_FIELD_NUMBER: _ClassVar[int]
    message_id: str
    offset: int
    status: _definition_pb2.Status
    transaction_id: str
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., message_id: _Optional[str] = ..., transaction_id: _Optional[str] = ..., offset: _Optional[int] = ...) -> None: ...

class TelemetryCommand(_message.Message):
    __slots__ = ["print_thread_stack_trace_command", "recover_orphaned_transaction_command", "settings", "status", "thread_stack_trace", "verify_message_command", "verify_message_result"]
    PRINT_THREAD_STACK_TRACE_COMMAND_FIELD_NUMBER: _ClassVar[int]
    RECOVER_ORPHANED_TRANSACTION_COMMAND_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    THREAD_STACK_TRACE_FIELD_NUMBER: _ClassVar[int]
    VERIFY_MESSAGE_COMMAND_FIELD_NUMBER: _ClassVar[int]
    VERIFY_MESSAGE_RESULT_FIELD_NUMBER: _ClassVar[int]
    print_thread_stack_trace_command: PrintThreadStackTraceCommand
    recover_orphaned_transaction_command: RecoverOrphanedTransactionCommand
    settings: _definition_pb2.Settings
    status: _definition_pb2.Status
    thread_stack_trace: ThreadStackTrace
    verify_message_command: VerifyMessageCommand
    verify_message_result: VerifyMessageResult
    def __init__(self, status: _Optional[_Union[_definition_pb2.Status, _Mapping]] = ..., settings: _Optional[_Union[_definition_pb2.Settings, _Mapping]] = ..., thread_stack_trace: _Optional[_Union[ThreadStackTrace, _Mapping]] = ..., verify_message_result: _Optional[_Union[VerifyMessageResult, _Mapping]] = ..., recover_orphaned_transaction_command: _Optional[_Union[RecoverOrphanedTransactionCommand, _Mapping]] = ..., print_thread_stack_trace_command: _Optional[_Union[PrintThreadStackTraceCommand, _Mapping]] = ..., verify_message_command: _Optional[_Union[VerifyMessageCommand, _Mapping]] = ...) -> None: ...

class ThreadStackTrace(_message.Message):
    __slots__ = ["nonce", "thread_stack_trace"]
    NONCE_FIELD_NUMBER: _ClassVar[int]
    THREAD_STACK_TRACE_FIELD_NUMBER: _ClassVar[int]
    nonce: str
    thread_stack_trace: str
    def __init__(self, nonce: _Optional[str] = ..., thread_stack_trace: _Optional[str] = ...) -> None: ...

class VerifyMessageCommand(_message.Message):
    __slots__ = ["message", "nonce"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    NONCE_FIELD_NUMBER: _ClassVar[int]
    message: _definition_pb2.Message
    nonce: str
    def __init__(self, nonce: _Optional[str] = ..., message: _Optional[_Union[_definition_pb2.Message, _Mapping]] = ...) -> None: ...

class VerifyMessageResult(_message.Message):
    __slots__ = ["nonce"]
    NONCE_FIELD_NUMBER: _ClassVar[int]
    nonce: str
    def __init__(self, nonce: _Optional[str] = ...) -> None: ...



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from rocketmq.protocol import service_pb2 as apache_dot_rocketmq_dot_v2_dot_service__pb2


class MessagingServiceStub(object):
    """For all the RPCs in MessagingService, the following error handling policies
    apply:

    If the request doesn't bear a valid authentication credential, return a
    response with common.status.code == `UNAUTHENTICATED`. If the authenticated
    user is not granted with sufficient permission to execute the requested
    operation, return a response with common.status.code == `PERMISSION_DENIED`.
    If the per-user-resource-based quota is exhausted, return a response with
    common.status.code == `RESOURCE_EXHAUSTED`. If any unexpected server-side
    errors raise, return a response with common.status.code == `INTERNAL`.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.QueryRoute = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/QueryRoute',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryRouteRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryRouteResponse.FromString,
                )
        self.Heartbeat = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/Heartbeat',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.HeartbeatRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.HeartbeatResponse.FromString,
                )
        self.SendMessage = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/SendMessage',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.SendMessageRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.SendMessageResponse.FromString,
                )
        self.QueryAssignment = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/QueryAssignment',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryAssignmentRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryAssignmentResponse.FromString,
                )
        self.ReceiveMessage = channel.unary_stream(
                '/apache.rocketmq.v2.MessagingService/ReceiveMessage',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ReceiveMessageRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ReceiveMessageResponse.FromString,
                )
        self.AckMessage = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/AckMessage',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.AckMessageRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.AckMessageResponse.FromString,
                )
        self.ForwardMessageToDeadLetterQueue = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/ForwardMessageToDeadLetterQueue',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ForwardMessageToDeadLetterQueueRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ForwardMessageToDeadLetterQueueResponse.FromString,
                )
        self.EndTransaction = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/EndTransaction',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.EndTransactionRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.EndTransactionResponse.FromString,
                )
        self.Telemetry = channel.stream_stream(
                '/apache.rocketmq.v2.MessagingService/Telemetry',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.TelemetryCommand.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.TelemetryCommand.FromString,
                )
        self.NotifyClientTermination = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/NotifyClientTermination',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.NotifyClientTerminationRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.NotifyClientTerminationResponse.FromString,
                )
        self.ChangeInvisibleDuration = channel.unary_unary(
                '/apache.rocketmq.v2.MessagingService/ChangeInvisibleDuration',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ChangeInvisibleDurationRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ChangeInvisibleDurationResponse.FromString,
                )


class MessagingServiceServicer(object):
    """For all the RPCs in MessagingService, the following error handling policies
    apply:

    If the request doesn't bear a valid authentication credential, return a
    response with common.status.code == `UNAUTHENTICATED`. If the authenticated
    user is not granted with sufficient permission to execute the requested
    operation, return a response with common.status.code == `PERMISSION_DENIED`.
    If the per-user-resource-based quota is exhausted, return a response with
    common.status.code == `RESOURCE_EXHAUSTED`. If any unexpected server-side
    errors raise, return a response with common.status.code == `INTERNAL`.
    """

    def QueryRoute(self, request, context):
        """Queries the route entries of the requested topic in the perspective of the
        given endpoints. On success, servers should return a collection of
        addressable message-queues. Note servers may return customized route
        entries based on endpoints provided.

        If the requested topic doesn't exist, returns `NOT_FOUND`.
        If the specific endpoints is empty, returns `INVALID_ARGUMENT`.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Heartbeat(self, request, context):
        """Producer or consumer sends HeartbeatRequest to servers periodically to
        keep-alive. Additionally, it also reports client-side configuration,
        including topic subscription, load-balancing group name, etc.

        Returns `OK` if success.

        If a client specifies a language that is not yet supported by servers,
        returns `INVALID_ARGUMENT`
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendMessage(self, request, context):
        """Delivers messages to brokers.
        Clients may further:
        1. Refine a message destination to message-queues which fulfills parts of
        FIFO semantic;
        2. Flag a message as transactional, which keeps it invisible to consumers
        until it commits;
        3. Time a message, making it invisible to consumers till specified
        time-point;
        4. And more...

        Returns message-id or transaction-id with status `OK` on success.

        If the destination topic doesn't exist, returns `NOT_FOUND`.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def QueryAssignment(self, request, context):
        """Queries the assigned route info of a topic for current consumer,
        the returned assignment result is decided by server-side load balancer.

        If the corresponding topic doesn't exist, returns `NOT_FOUND`.
        If the specific endpoints is empty, returns `INVALID_ARGUMENT`.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReceiveMessage(self, request, context):
        """Receives messages from the server in batch manner, returns a set of
        messages if success. The received messages should be acked or redelivered
        after processed.

        If the pending concurrent receive requests exceed the quota of the given
        consumer group, returns `UNAVAILABLE`. If the upstream store server hangs,
        return `DEADLINE_EXCEEDED` in a timely manner. If the corresponding topic
        or consumer group doesn't exist, returns `NOT_FOUND`. If there is no new
        message in the specific topic, returns `OK` with an empty message set.
        Please note that client may suffer from false empty responses.

        If failed to receive message from remote, server must return only one
        `ReceiveMessageResponse` as the reply to the request, whose `Status` indicates
        the specific reason of failure, otherwise, the reply is considered successful.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AckMessage(self, request, context):
        """Acknowledges the message associated with the `receipt_handle` or `offset`
        in the `AckMessageRequest`, it means the message has been successfully
        processed. Returns `OK` if the message server remove the relevant message
        successfully.

        If the given receipt_handle is illegal or out of date, returns
        `INVALID_ARGUMENT`.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ForwardMessageToDeadLetterQueue(self, request, context):
        """Forwards one message to dead letter queue if the max delivery attempts is
        exceeded by this message at client-side, return `OK` if success.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EndTransaction(self, request, context):
        """Commits or rollback one transactional message.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Telemetry(self, request_iterator, context):
        """Once a client starts, it would immediately establishes bi-lateral stream
        RPCs with brokers, reporting its settings as the initiative command.

        When servers have need of inspecting client status, they would issue
        telemetry commands to clients. After executing received instructions,
        clients shall report command execution results through client-side streams.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NotifyClientTermination(self, request, context):
        """Notify the server that the client is terminated.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ChangeInvisibleDuration(self, request, context):
        """Once a message is retrieved from consume queue on behalf of the group, it
        will be kept invisible to other clients of the same group for a period of
        time. The message is supposed to be processed within the invisible
        duration. If the client, which is in charge of the invisible message, is
        not capable of processing the message timely, it may use
        ChangeInvisibleDuration to lengthen invisible duration.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MessagingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'QueryRoute': grpc.unary_unary_rpc_method_handler(
                    servicer.QueryRoute,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryRouteRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryRouteResponse.SerializeToString,
            ),
            'Heartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.Heartbeat,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.HeartbeatRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.HeartbeatResponse.SerializeToString,
            ),
            'SendMessage': grpc.unary_unary_rpc_method_handler(
                    servicer.SendMessage,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.SendMessageRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.SendMessageResponse.SerializeToString,
            ),
            'QueryAssignment': grpc.unary_unary_rpc_method_handler(
                    servicer.QueryAssignment,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryAssignmentRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryAssignmentResponse.SerializeToString,
            ),
            'ReceiveMessage': grpc.unary_stream_rpc_method_handler(
                    servicer.ReceiveMessage,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ReceiveMessageRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ReceiveMessageResponse.SerializeToString,
            ),
            'AckMessage': grpc.unary_unary_rpc_method_handler(
                    servicer.AckMessage,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.AckMessageRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.AckMessageResponse.SerializeToString,
            ),
            'ForwardMessageToDeadLetterQueue': grpc.unary_unary_rpc_method_handler(
                    servicer.ForwardMessageToDeadLetterQueue,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ForwardMessageToDeadLetterQueueRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ForwardMessageToDeadLetterQueueResponse.SerializeToString,
            ),
            'EndTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.EndTransaction,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.EndTransactionRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.EndTransactionResponse.SerializeToString,
            ),
            'Telemetry': grpc.stream_stream_rpc_method_handler(
                    servicer.Telemetry,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.TelemetryCommand.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.TelemetryCommand.SerializeToString,
            ),
            'NotifyClientTermination': grpc.unary_unary_rpc_method_handler(
                    servicer.NotifyClientTermination,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.NotifyClientTerminationRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.NotifyClientTerminationResponse.SerializeToString,
            ),
            'ChangeInvisibleDuration': grpc.unary_unary_rpc_method_handler(
                    servicer.ChangeInvisibleDuration,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ChangeInvisibleDurationRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_service__pb2.ChangeInvisibleDurationResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'apache.rocketmq.v2.MessagingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class MessagingService(object):
    """For all the RPCs in MessagingService, the following error handling policies
    apply:

    If the request doesn't bear a valid authentication credential, return a
    response with common.status.code == `UNAUTHENTICATED`. If the authenticated
    user is not granted with sufficient permission to execute the requested
    operation, return a response with common.status.code == `PERMISSION_DENIED`.
    If the per-user-resource-based quota is exhausted, return a response with
    common.status.code == `RESOURCE_EXHAUSTED`. If any unexpected server-side
    errors raise, return a response with common.status.code == `INTERNAL`.
    """

    @staticmethod
    def QueryRoute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/QueryRoute',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryRouteRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryRouteResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Heartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/Heartbeat',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.HeartbeatRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.HeartbeatResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendMessage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/SendMessage',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.SendMessageRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.SendMessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def QueryAssignment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/QueryAssignment',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryAssignmentRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.QueryAssignmentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReceiveMessage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/apache.rocketmq.v2.MessagingService/ReceiveMessage',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.ReceiveMessageRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.ReceiveMessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AckMessage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/AckMessage',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.AckMessageRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.AckMessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ForwardMessageToDeadLetterQueue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/ForwardMessageToDeadLetterQueue',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.ForwardMessageToDeadLetterQueueRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.ForwardMessageToDeadLetterQueueResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def EndTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/EndTransaction',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.EndTransactionRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.EndTransactionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Telemetry(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/apache.rocketmq.v2.MessagingService/Telemetry',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.TelemetryCommand.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.TelemetryCommand.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def NotifyClientTermination(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/NotifyClientTermination',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.NotifyClientTerminationRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.NotifyClientTerminationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ChangeInvisibleDuration(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.MessagingService/ChangeInvisibleDuration',
            apache_dot_rocketmq_dot_v2_dot_service__pb2.ChangeInvisibleDurationRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_service__pb2.ChangeInvisibleDurationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from rocketmq.status_checker import StatusChecker
from rocketmq.log import logger
from rocketmq.message_id import MessageId
from rocketmq.protocol.definition_pb2 import Code as ProtoCode


class SendReceipt:
    def __init__(self, message_id: MessageId, transaction_id, message_queue):
        self.message_id = message_id
        self.transaction_id = transaction_id
        self.message_queue = message_queue

    @property
    def endpoints(self):
        return self.message_queue.broker.endpoints

    def __str__(self):
        return f'MessageId: {self.message_id}'

    @staticmethod
    def process_send_message_response(mq, invocation):
        status = invocation.status
        for entry in invocation.entries:
            if entry.status.code == ProtoCode.OK:
                status = entry.status
        logger.debug(status)
        # May throw exception.
        # StatusChecker.check(status, invocation.request, invocation.request_id)
        return [SendReceipt(entry.message_id, entry.transaction_id, mq) for entry in invocation.entries]



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rocketmq.rpc_client import Endpoints
from rocketmq.session_credentials import SessionCredentialsProvider


class ClientConfig:
    """Client configuration class which holds the settings for a client.
    The settings include endpoint configurations, session credential provider and SSL settings.
    An instance of this class is used to setup the client with necessary configurations.
    """

    def __init__(
        self,
        endpoints: Endpoints,
        session_credentials_provider: SessionCredentialsProvider,
        ssl_enabled: bool,
    ):
        #: The endpoints for the client to connect to.
        self.__endpoints = endpoints

        #: The session credentials provider to authenticate the client.
        self.__session_credentials_provider = session_credentials_provider

        #: A flag indicating if SSL is enabled for the client.
        self.__ssl_enabled = ssl_enabled

        #: The request timeout for the client in seconds.
        self.request_timeout = 10

    @property
    def session_credentials_provider(self) -> SessionCredentialsProvider:
        """The session credentials provider for the client.

        :return: the session credentials provider
        """
        return self.__session_credentials_provider

    @property
    def endpoints(self) -> Endpoints:
        """The endpoints for the client to connect to.

        :return: the endpoints
        """
        return self.__endpoints

    @property
    def ssl_enabled(self) -> bool:
        """A flag indicating if SSL is enabled for the client.

        :return: True if SSL is enabled, False otherwise
        """
        return self.__ssl_enabled



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ChangeLogLevelRequest(_message.Message):
    __slots__ = ["level"]
    class Level(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    DEBUG: ChangeLogLevelRequest.Level
    ERROR: ChangeLogLevelRequest.Level
    INFO: ChangeLogLevelRequest.Level
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    TRACE: ChangeLogLevelRequest.Level
    WARN: ChangeLogLevelRequest.Level
    level: ChangeLogLevelRequest.Level
    def __init__(self, level: _Optional[_Union[ChangeLogLevelRequest.Level, str]] = ...) -> None: ...

class ChangeLogLevelResponse(_message.Message):
    __slots__ = ["remark"]
    REMARK_FIELD_NUMBER: _ClassVar[int]
    remark: str
    def __init__(self, remark: _Optional[str] = ...) -> None: ...



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

ADDRESS_SCHEME_UNSPECIFIED: AddressScheme
BAD_REQUEST: Code
CLIENT_ID_REQUIRED: Code
CLIENT_TYPE_UNSPECIFIED: ClientType
CODE_UNSPECIFIED: Code
COMMIT: TransactionResolution
CONSUMER_GROUP_NOT_FOUND: Code
CPP: Language
CRC32: DigestType
DART: Language
DELAY: MessageType
DESCRIPTOR: _descriptor.FileDescriptor
DIGEST_TYPE_UNSPECIFIED: DigestType
DOMAIN_NAME: AddressScheme
DOT_NET: Language
ENCODING_UNSPECIFIED: Encoding
FAILED_TO_CONSUME_MESSAGE: Code
FIFO: MessageType
FILTER_TYPE_UNSPECIFIED: FilterType
FORBIDDEN: Code
GOLANG: Language
GZIP: Encoding
HA_NOT_AVAILABLE: Code
IDENTITY: Encoding
ILLEGAL_ACCESS_POINT: Code
ILLEGAL_CONSUMER_GROUP: Code
ILLEGAL_DELIVERY_TIME: Code
ILLEGAL_FILTER_EXPRESSION: Code
ILLEGAL_INVISIBLE_TIME: Code
ILLEGAL_MESSAGE_GROUP: Code
ILLEGAL_MESSAGE_ID: Code
ILLEGAL_MESSAGE_KEY: Code
ILLEGAL_MESSAGE_PROPERTY_KEY: Code
ILLEGAL_MESSAGE_TAG: Code
ILLEGAL_POLLING_TIME: Code
ILLEGAL_TOPIC: Code
INTERNAL_ERROR: Code
INTERNAL_SERVER_ERROR: Code
INVALID_RECEIPT_HANDLE: Code
INVALID_TRANSACTION_ID: Code
IPv4: AddressScheme
IPv6: AddressScheme
JAVA: Language
KOTLIN: Language
LANGUAGE_UNSPECIFIED: Language
MASTER_PERSISTENCE_TIMEOUT: Code
MD5: DigestType
MESSAGE_BODY_TOO_LARGE: Code
MESSAGE_CORRUPTED: Code
MESSAGE_NOT_FOUND: Code
MESSAGE_PROPERTIES_TOO_LARGE: Code
MESSAGE_PROPERTY_CONFLICT_WITH_TYPE: Code
MESSAGE_TYPE_UNSPECIFIED: MessageType
MULTIPLE_RESULTS: Code
NODE_JS: Language
NONE: Permission
NORMAL: MessageType
NOT_FOUND: Code
NOT_IMPLEMENTED: Code
OBJECTIVE_C: Language
OK: Code
PAYLOAD_TOO_LARGE: Code
PAYMENT_REQUIRED: Code
PERMISSION_UNSPECIFIED: Permission
PHP: Language
PRECONDITION_FAILED: Code
PRODUCER: ClientType
PROXY_TIMEOUT: Code
PUSH_CONSUMER: ClientType
PYTHON: Language
READ: Permission
READ_WRITE: Permission
REQUEST_HEADER_FIELDS_TOO_LARGE: Code
REQUEST_TIMEOUT: Code
ROLLBACK: TransactionResolution
RUBY: Language
RUST: Language
SHA1: DigestType
SIMPLE_CONSUMER: ClientType
SLAVE_PERSISTENCE_TIMEOUT: Code
SOURCE_CLIENT: TransactionSource
SOURCE_SERVER_CHECK: TransactionSource
SOURCE_UNSPECIFIED: TransactionSource
SQL: FilterType
TAG: FilterType
TOO_MANY_REQUESTS: Code
TOPIC_NOT_FOUND: Code
TRANSACTION: MessageType
TRANSACTION_RESOLUTION_UNSPECIFIED: TransactionResolution
UNAUTHORIZED: Code
UNRECOGNIZED_CLIENT_TYPE: Code
UNSUPPORTED: Code
VERIFY_FIFO_MESSAGE_UNSUPPORTED: Code
VERSION_UNSUPPORTED: Code
WRITE: Permission

class Address(_message.Message):
    __slots__ = ["host", "port"]
    HOST_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    host: str
    port: int
    def __init__(self, host: _Optional[str] = ..., port: _Optional[int] = ...) -> None: ...

class Assignment(_message.Message):
    __slots__ = ["message_queue"]
    MESSAGE_QUEUE_FIELD_NUMBER: _ClassVar[int]
    message_queue: MessageQueue
    def __init__(self, message_queue: _Optional[_Union[MessageQueue, _Mapping]] = ...) -> None: ...

class Broker(_message.Message):
    __slots__ = ["endpoints", "id", "name"]
    ENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    endpoints: Endpoints
    id: int
    name: str
    def __init__(self, name: _Optional[str] = ..., id: _Optional[int] = ..., endpoints: _Optional[_Union[Endpoints, _Mapping]] = ...) -> None: ...

class CustomizedBackoff(_message.Message):
    __slots__ = ["next"]
    NEXT_FIELD_NUMBER: _ClassVar[int]
    next: _containers.RepeatedCompositeFieldContainer[_duration_pb2.Duration]
    def __init__(self, next: _Optional[_Iterable[_Union[_duration_pb2.Duration, _Mapping]]] = ...) -> None: ...

class DeadLetterQueue(_message.Message):
    __slots__ = ["message_id", "topic"]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    message_id: str
    topic: str
    def __init__(self, topic: _Optional[str] = ..., message_id: _Optional[str] = ...) -> None: ...

class Digest(_message.Message):
    __slots__ = ["checksum", "type"]
    CHECKSUM_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    checksum: str
    type: DigestType
    def __init__(self, type: _Optional[_Union[DigestType, str]] = ..., checksum: _Optional[str] = ...) -> None: ...

class Endpoints(_message.Message):
    __slots__ = ["addresses", "scheme"]
    ADDRESSES_FIELD_NUMBER: _ClassVar[int]
    SCHEME_FIELD_NUMBER: _ClassVar[int]
    addresses: _containers.RepeatedCompositeFieldContainer[Address]
    scheme: AddressScheme
    def __init__(self, scheme: _Optional[_Union[AddressScheme, str]] = ..., addresses: _Optional[_Iterable[_Union[Address, _Mapping]]] = ...) -> None: ...

class ExponentialBackoff(_message.Message):
    __slots__ = ["initial", "max", "multiplier"]
    INITIAL_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    MULTIPLIER_FIELD_NUMBER: _ClassVar[int]
    initial: _duration_pb2.Duration
    max: _duration_pb2.Duration
    multiplier: float
    def __init__(self, initial: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., max: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., multiplier: _Optional[float] = ...) -> None: ...

class FilterExpression(_message.Message):
    __slots__ = ["expression", "type"]
    EXPRESSION_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    expression: str
    type: FilterType
    def __init__(self, type: _Optional[_Union[FilterType, str]] = ..., expression: _Optional[str] = ...) -> None: ...

class Message(_message.Message):
    __slots__ = ["body", "system_properties", "topic", "user_properties"]
    class UserPropertiesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    BODY_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    USER_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    body: bytes
    system_properties: SystemProperties
    topic: Resource
    user_properties: _containers.ScalarMap[str, str]
    def __init__(self, topic: _Optional[_Union[Resource, _Mapping]] = ..., user_properties: _Optional[_Mapping[str, str]] = ..., system_properties: _Optional[_Union[SystemProperties, _Mapping]] = ..., body: _Optional[bytes] = ...) -> None: ...

class MessageQueue(_message.Message):
    __slots__ = ["accept_message_types", "broker", "id", "permission", "topic"]
    ACCEPT_MESSAGE_TYPES_FIELD_NUMBER: _ClassVar[int]
    BROKER_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    PERMISSION_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    accept_message_types: _containers.RepeatedScalarFieldContainer[MessageType]
    broker: Broker
    id: int
    permission: Permission
    topic: Resource
    def __init__(self, topic: _Optional[_Union[Resource, _Mapping]] = ..., id: _Optional[int] = ..., permission: _Optional[_Union[Permission, str]] = ..., broker: _Optional[_Union[Broker, _Mapping]] = ..., accept_message_types: _Optional[_Iterable[_Union[MessageType, str]]] = ...) -> None: ...

class Metric(_message.Message):
    __slots__ = ["endpoints", "on"]
    ENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    ON_FIELD_NUMBER: _ClassVar[int]
    endpoints: Endpoints
    on: bool
    def __init__(self, on: bool = ..., endpoints: _Optional[_Union[Endpoints, _Mapping]] = ...) -> None: ...

class Publishing(_message.Message):
    __slots__ = ["max_body_size", "topics", "validate_message_type"]
    MAX_BODY_SIZE_FIELD_NUMBER: _ClassVar[int]
    TOPICS_FIELD_NUMBER: _ClassVar[int]
    VALIDATE_MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    max_body_size: int
    topics: _containers.RepeatedCompositeFieldContainer[Resource]
    validate_message_type: bool
    def __init__(self, topics: _Optional[_Iterable[_Union[Resource, _Mapping]]] = ..., max_body_size: _Optional[int] = ..., validate_message_type: bool = ...) -> None: ...

class Resource(_message.Message):
    __slots__ = ["name", "resource_namespace"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    name: str
    resource_namespace: str
    def __init__(self, resource_namespace: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class RetryPolicy(_message.Message):
    __slots__ = ["customized_backoff", "exponential_backoff", "max_attempts"]
    CUSTOMIZED_BACKOFF_FIELD_NUMBER: _ClassVar[int]
    EXPONENTIAL_BACKOFF_FIELD_NUMBER: _ClassVar[int]
    MAX_ATTEMPTS_FIELD_NUMBER: _ClassVar[int]
    customized_backoff: CustomizedBackoff
    exponential_backoff: ExponentialBackoff
    max_attempts: int
    def __init__(self, max_attempts: _Optional[int] = ..., exponential_backoff: _Optional[_Union[ExponentialBackoff, _Mapping]] = ..., customized_backoff: _Optional[_Union[CustomizedBackoff, _Mapping]] = ...) -> None: ...

class Settings(_message.Message):
    __slots__ = ["access_point", "backoff_policy", "client_type", "metric", "publishing", "request_timeout", "subscription", "user_agent"]
    ACCESS_POINT_FIELD_NUMBER: _ClassVar[int]
    BACKOFF_POLICY_FIELD_NUMBER: _ClassVar[int]
    CLIENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    METRIC_FIELD_NUMBER: _ClassVar[int]
    PUBLISHING_FIELD_NUMBER: _ClassVar[int]
    REQUEST_TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTION_FIELD_NUMBER: _ClassVar[int]
    USER_AGENT_FIELD_NUMBER: _ClassVar[int]
    access_point: Endpoints
    backoff_policy: RetryPolicy
    client_type: ClientType
    metric: Metric
    publishing: Publishing
    request_timeout: _duration_pb2.Duration
    subscription: Subscription
    user_agent: UA
    def __init__(self, client_type: _Optional[_Union[ClientType, str]] = ..., access_point: _Optional[_Union[Endpoints, _Mapping]] = ..., backoff_policy: _Optional[_Union[RetryPolicy, _Mapping]] = ..., request_timeout: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., publishing: _Optional[_Union[Publishing, _Mapping]] = ..., subscription: _Optional[_Union[Subscription, _Mapping]] = ..., user_agent: _Optional[_Union[UA, _Mapping]] = ..., metric: _Optional[_Union[Metric, _Mapping]] = ...) -> None: ...

class Status(_message.Message):
    __slots__ = ["code", "message"]
    CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    code: Code
    message: str
    def __init__(self, code: _Optional[_Union[Code, str]] = ..., message: _Optional[str] = ...) -> None: ...

class Subscription(_message.Message):
    __slots__ = ["fifo", "group", "long_polling_timeout", "receive_batch_size", "subscriptions"]
    FIFO_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    LONG_POLLING_TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    RECEIVE_BATCH_SIZE_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
    fifo: bool
    group: Resource
    long_polling_timeout: _duration_pb2.Duration
    receive_batch_size: int
    subscriptions: _containers.RepeatedCompositeFieldContainer[SubscriptionEntry]
    def __init__(self, group: _Optional[_Union[Resource, _Mapping]] = ..., subscriptions: _Optional[_Iterable[_Union[SubscriptionEntry, _Mapping]]] = ..., fifo: bool = ..., receive_batch_size: _Optional[int] = ..., long_polling_timeout: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class SubscriptionEntry(_message.Message):
    __slots__ = ["expression", "topic"]
    EXPRESSION_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    expression: FilterExpression
    topic: Resource
    def __init__(self, topic: _Optional[_Union[Resource, _Mapping]] = ..., expression: _Optional[_Union[FilterExpression, _Mapping]] = ...) -> None: ...

class SystemProperties(_message.Message):
    __slots__ = ["body_digest", "body_encoding", "born_host", "born_timestamp", "dead_letter_queue", "delivery_attempt", "delivery_timestamp", "invisible_duration", "keys", "message_group", "message_id", "message_type", "orphaned_transaction_recovery_duration", "queue_id", "queue_offset", "receipt_handle", "store_host", "store_timestamp", "tag", "trace_context"]
    BODY_DIGEST_FIELD_NUMBER: _ClassVar[int]
    BODY_ENCODING_FIELD_NUMBER: _ClassVar[int]
    BORN_HOST_FIELD_NUMBER: _ClassVar[int]
    BORN_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DEAD_LETTER_QUEUE_FIELD_NUMBER: _ClassVar[int]
    DELIVERY_ATTEMPT_FIELD_NUMBER: _ClassVar[int]
    DELIVERY_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    INVISIBLE_DURATION_FIELD_NUMBER: _ClassVar[int]
    KEYS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_GROUP_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    ORPHANED_TRANSACTION_RECOVERY_DURATION_FIELD_NUMBER: _ClassVar[int]
    QUEUE_ID_FIELD_NUMBER: _ClassVar[int]
    QUEUE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    RECEIPT_HANDLE_FIELD_NUMBER: _ClassVar[int]
    STORE_HOST_FIELD_NUMBER: _ClassVar[int]
    STORE_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    TAG_FIELD_NUMBER: _ClassVar[int]
    TRACE_CONTEXT_FIELD_NUMBER: _ClassVar[int]
    body_digest: Digest
    body_encoding: Encoding
    born_host: str
    born_timestamp: _timestamp_pb2.Timestamp
    dead_letter_queue: DeadLetterQueue
    delivery_attempt: int
    delivery_timestamp: _timestamp_pb2.Timestamp
    invisible_duration: _duration_pb2.Duration
    keys: _containers.RepeatedScalarFieldContainer[str]
    message_group: str
    message_id: str
    message_type: MessageType
    orphaned_transaction_recovery_duration: _duration_pb2.Duration
    queue_id: int
    queue_offset: int
    receipt_handle: str
    store_host: str
    store_timestamp: _timestamp_pb2.Timestamp
    tag: str
    trace_context: str
    def __init__(self, tag: _Optional[str] = ..., keys: _Optional[_Iterable[str]] = ..., message_id: _Optional[str] = ..., body_digest: _Optional[_Union[Digest, _Mapping]] = ..., body_encoding: _Optional[_Union[Encoding, str]] = ..., message_type: _Optional[_Union[MessageType, str]] = ..., born_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., born_host: _Optional[str] = ..., store_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., store_host: _Optional[str] = ..., delivery_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., receipt_handle: _Optional[str] = ..., queue_id: _Optional[int] = ..., queue_offset: _Optional[int] = ..., invisible_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., delivery_attempt: _Optional[int] = ..., message_group: _Optional[str] = ..., trace_context: _Optional[str] = ..., orphaned_transaction_recovery_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., dead_letter_queue: _Optional[_Union[DeadLetterQueue, _Mapping]] = ...) -> None: ...

class UA(_message.Message):
    __slots__ = ["hostname", "language", "platform", "version"]
    HOSTNAME_FIELD_NUMBER: _ClassVar[int]
    LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    PLATFORM_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    hostname: str
    language: Language
    platform: str
    version: str
    def __init__(self, language: _Optional[_Union[Language, str]] = ..., version: _Optional[str] = ..., platform: _Optional[str] = ..., hostname: _Optional[str] = ...) -> None: ...

class TransactionResolution(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class TransactionSource(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class Permission(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class FilterType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class AddressScheme(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MessageType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class DigestType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ClientType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class Encoding(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class Code(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class Language(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from protocol import admin_pb2 as apache_dot_rocketmq_dot_v2_dot_admin__pb2


class AdminStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ChangeLogLevel = channel.unary_unary(
                '/apache.rocketmq.v2.Admin/ChangeLogLevel',
                request_serializer=apache_dot_rocketmq_dot_v2_dot_admin__pb2.ChangeLogLevelRequest.SerializeToString,
                response_deserializer=apache_dot_rocketmq_dot_v2_dot_admin__pb2.ChangeLogLevelResponse.FromString,
                )


class AdminServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ChangeLogLevel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AdminServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ChangeLogLevel': grpc.unary_unary_rpc_method_handler(
                    servicer.ChangeLogLevel,
                    request_deserializer=apache_dot_rocketmq_dot_v2_dot_admin__pb2.ChangeLogLevelRequest.FromString,
                    response_serializer=apache_dot_rocketmq_dot_v2_dot_admin__pb2.ChangeLogLevelResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'apache.rocketmq.v2.Admin', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Admin(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ChangeLogLevel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/apache.rocketmq.v2.Admin/ChangeLogLevel',
            apache_dot_rocketmq_dot_v2_dot_admin__pb2.ChangeLogLevelRequest.SerializeToString,
            apache_dot_rocketmq_dot_v2_dot_admin__pb2.ChangeLogLevelResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
